import sys

import kaggle
import zipfile
import os
import pandas as pd


class KaggleDatasetDownloader:
    def __init__(self, dataset, file_name, download_path):
        self.dataset = dataset
        self.file_name = file_name
        self.download_path = download_path

    def authenticate_and_download(self):
        kaggle.api.authenticate()
        kaggle.api.dataset_download_file(self.dataset, self.file_name, path=self.download_path)
        self.extract_if_needed()

    def extract_if_needed(self):
        zip_path = os.path.join(self.download_path, f'{self.file_name}.zip')
        if os.path.exists(zip_path):
            with zipfile.ZipFile(zip_path, 'r') as zip_ref:
                zip_ref.extractall(self.download_path)
            os.remove(zip_path)


class DatasetProcessor:
    def __init__(self, dataset_path):
        self.dataset = pd.read_csv(dataset_path)

    def filter_and_process(self):
        filtered_positive = self.dataset[self.dataset['rating'] > 3].copy()
        filtered_negative = self.dataset[self.dataset['rating'] < 3].copy()

        filtered_positive = self.clean_rubrics(filtered_positive)
        filtered_negative = self.clean_rubrics(filtered_negative)

        rename_mapping = {
            'Пиццерия': 'Быстрое питание', 'Ресторан': 'Быстрое питание',
            'Кафе': 'Быстрое питание', 'Магазин продуктов': 'Супермаркет',
            'Аптека': 'Медцентр, клиника', 'Стоматологическая клиника': 'Медцентр, клиника'
        }

        filtered_positive = self.rename_rubrics(filtered_positive, rename_mapping)
        filtered_negative = self.rename_rubrics(filtered_negative, rename_mapping)

        unique_rubrics_positive = self.get_top_rubrics(filtered_positive)
        unique_rubrics_negative = self.get_top_rubrics(filtered_negative)

        filtered_positive = self.filter_by_rubrics(filtered_positive, unique_rubrics_positive)
        filtered_negative = self.filter_by_rubrics(filtered_negative, unique_rubrics_negative)

        filtered_positive = self.limit_rows_per_rubric(filtered_positive, 15000)
        filtered_negative = self.limit_rows_per_rubric(filtered_negative, 1500)

        self.print_rubric_counts(filtered_positive, unique_rubrics_positive)
        self.print_rubric_counts(filtered_negative, unique_rubrics_negative)

        return filtered_positive, filtered_negative, unique_rubrics_positive, unique_rubrics_negative

    @staticmethod
    def clean_rubrics(dataset):
        dataset.loc[:, 'rubrics'] = dataset['rubrics'].apply(lambda x: x.split(';')[0] if pd.notnull(x) else x)
        return dataset

    @staticmethod
    def rename_rubrics(dataset, rename_mapping):
        dataset.loc[:, 'rubrics'] = dataset['rubrics'].replace(rename_mapping)
        return dataset

    @staticmethod
    def get_top_rubrics(dataset, top_n=10):
        return set(dataset['rubrics'].value_counts().head(top_n).index.tolist())

    @staticmethod
    def filter_by_rubrics(dataset, rubrics):
        return dataset[dataset['rubrics'].isin(rubrics)]

    @staticmethod
    def limit_rows_per_rubric(dataset, limit):
        return dataset.groupby('rubrics').head(limit).reset_index(drop=True)

    @staticmethod
    def print_rubric_counts(dataset, rubrics):
        for rubric in rubrics:
            count = dataset[dataset['rubrics'] == rubric].shape[0]
            print(f"Количество строк с rubrics '{rubric}': {count}")



class FileSaver:
    @staticmethod
    def save_rubrics_to_file(rubrics, file_path):
        with open(file_path, 'w', encoding='utf-8') as file:
            for rubric in rubrics:
                file.write(f"{rubric}\n")

    @staticmethod
    def save_dataset_to_csv(dataset, file_path):
        dataset.to_csv(file_path, index=False)


def main():
    dataset = 'kyakovlev/yandex-geo-reviews-dataset-2023'
    file_name = 'geo-reviews-dataset-2023.csv'
    download_path = '../dataset'
    dataset_path = os.path.join(download_path, file_name)

    # Скачивание и распаковка датасета
    downloader = KaggleDatasetDownloader(dataset, file_name, download_path)
    downloader.authenticate_and_download()

    # Обработка датасета
    processor = DatasetProcessor(dataset_path)
    filtered_positive, filtered_negative, unique_rubrics_positive, unique_rubrics_negative = processor.filter_and_process()

    # Сохранение результатов
    FileSaver.save_rubrics_to_file(unique_rubrics_positive, 'class/class_positive.txt')
    FileSaver.save_rubrics_to_file(unique_rubrics_negative, 'class/class_negative.txt')
    FileSaver.save_dataset_to_csv(filtered_positive, 'filtered/filtered_dataset_positive.csv')
    FileSaver.save_dataset_to_csv(filtered_negative, 'filtered/filtered_dataset_negative.csv')


if __name__ == "__main__":
    main()
