import pickle
import pandas as pd
from keras import Sequential
from keras.src.callbacks import ModelCheckpoint
from keras.src.utils import pad_sequences
from matplotlib import pyplot as plt
from tensorflow.keras.layers import Dense, Embedding, SpatialDropout1D, LSTM, GRU, Dropout, Conv1D, GlobalMaxPooling1D
from tensorflow.keras import utils
from sklearn.preprocessing import LabelEncoder

# Определение констант
NUM_WORDS = 10000
MAX_REVIEWS_LEN = 90
NB_CLASSES = 10

class DataProcessor:
    def __init__(self, tokenizer_path, class_names_path, dataset_path):
        self.tokenizer = self.load_tokenizer(tokenizer_path)
        self.class_names = self.load_class_names(class_names_path)
        self.dataset = self.load_dataset(dataset_path)
        self.label_encoder = LabelEncoder()

    @staticmethod
    def load_tokenizer(tokenizer_path):
        with open(tokenizer_path, 'rb') as handle:
            return pickle.load(handle)

    @staticmethod
    def load_class_names(class_names_path):
        with open(class_names_path, 'r', encoding='utf-8') as file:
            return [line.strip() for line in file.readlines()]

    @staticmethod
    def load_dataset(dataset_path):
        train = pd.read_csv(dataset_path)
        train.drop(['address', 'name_ru', 'rating'], axis=1, inplace=True)
        return train

    def preprocess_data(self):
        reviews = self.dataset['text']
        encoded_labels = self.encode_labels(self.dataset['rubrics'])
        y_train = utils.to_categorical(encoded_labels, NB_CLASSES)
        sequences = self.tokenizer.texts_to_sequences(reviews)
        x_train = pad_sequences(sequences, maxlen=MAX_REVIEWS_LEN)
        return x_train, y_train

    def encode_labels(self, labels):
        self.label_encoder.fit(self.class_names)
        return self.label_encoder.transform(labels)

class CNNModelNegative:
    def __init__(self, num_words, max_reviews_len, nb_classes):
        self.num_words = num_words
        self.max_reviews_len = max_reviews_len
        self.nb_classes = nb_classes
        self.model = self.build_model()

    def build_model(self):
        model = Sequential()
        model.add(Embedding(self.num_words, 128, input_length=self.max_reviews_len))
        model.add(Conv1D(128, 5, activation='relu'))
        model.add(GlobalMaxPooling1D())
        model.add(Dense(64, activation='relu'))
        model.add(Dense(self.nb_classes, activation='softmax'))

        model.compile(optimizer='adam',
                      loss='categorical_crossentropy',
                      metrics=['accuracy'])
        return model

    def train(self, x_train, y_train, epochs, batch_size, validation_split, model_save_path):
        checkpoint_callback = ModelCheckpoint(model_save_path, monitor='val_accuracy', save_best_only=True, verbose=1)
        history = self.model.fit(x_train, y_train, epochs=epochs, batch_size=batch_size, validation_split=validation_split, callbacks=[checkpoint_callback])
        return history

class GRUModelNegative:
    def __init__(self, num_words, max_reviews_len, nb_classes):
        self.num_words = num_words
        self.max_reviews_len = max_reviews_len
        self.nb_classes = nb_classes
        self.model = self.build_model()

    def build_model(self):
        model = Sequential()
        model.add(Embedding(self.num_words, 128, input_length=self.max_reviews_len))
        model.add(GRU(128, return_sequences=True))
        model.add(Dropout(0.25))
        model.add(GRU(64))
        model.add(Dense(self.nb_classes, activation='softmax'))
        model.compile(optimizer='adam', loss='categorical_crossentropy', metrics=['accuracy'])
        return model

    def train(self, x_train, y_train, epochs, batch_size, validation_split, model_save_path):
        checkpoint_callback = ModelCheckpoint(model_save_path, monitor='val_accuracy', save_best_only=True, verbose=1)
        history = self.model.fit(x_train, y_train, epochs=epochs, batch_size=batch_size, validation_split=validation_split, callbacks=[checkpoint_callback])
        return history
class LSTMModelNegative:
    def __init__(self, num_words, max_reviews_len, nb_classes):
        self.num_words = num_words
        self.max_reviews_len = max_reviews_len
        self.nb_classes = nb_classes
        self.model = self.build_model()

    def build_model(self):
        model = Sequential()
        model.add(Embedding(self.num_words, 128, input_length=self.max_reviews_len))
        model.add(LSTM(64, dropout=0.2, recurrent_dropout=0.2))
        model.add(Dense(self.nb_classes, activation='softmax'))
        model.compile(optimizer='adam', loss='categorical_crossentropy', metrics=['accuracy'])
        return model

    def train(self, x_train, y_train, epochs, batch_size, validation_split, model_save_path):
        checkpoint_callback = ModelCheckpoint(model_save_path, monitor='val_accuracy', save_best_only=True, verbose=1)
        history = self.model.fit(x_train, y_train, epochs=epochs, batch_size=batch_size, validation_split=validation_split, callbacks=[checkpoint_callback])
        return history

class LSTMModelPositive:
    def __init__(self, num_words, max_reviews_len, nb_classes):
        self.num_words = num_words
        self.max_reviews_len = max_reviews_len
        self.nb_classes = nb_classes
        self.model = self.build_model()

    def build_model(self):
        model = Sequential()
        model.add(Embedding(self.num_words, 256, input_length=self.max_reviews_len))
        model.add(SpatialDropout1D(0.2))
        model.add(LSTM(128, dropout=0.2, recurrent_dropout=0.2))
        model.add(Dense(self.nb_classes, activation='softmax'))
        model.compile(optimizer='adam', loss='categorical_crossentropy', metrics=['accuracy'])
        return model

    def train(self, x_train, y_train, epochs, batch_size, validation_split, model_save_path):
        checkpoint_callback = ModelCheckpoint(model_save_path, monitor='val_accuracy', save_best_only=True, verbose=1)
        history = self.model.fit(x_train, y_train, epochs=epochs, batch_size=batch_size, validation_split=validation_split, callbacks=[checkpoint_callback])
        return history

class GRUModelPositive:
    def __init__(self, num_words, max_reviews_len, nb_classes):
        self.num_words = num_words
        self.max_reviews_len = max_reviews_len
        self.nb_classes = nb_classes
        self.model = self.build_model()

    def build_model(self):
        model = Sequential()
        model.add(Embedding(self.num_words, 256, input_length=self.max_reviews_len))
        model.add(GRU(256, return_sequences=True))
        model.add(Dropout(0.25))
        model.add(GRU(128))
        model.add(Dense(self.nb_classes, activation='softmax'))
        model.compile(optimizer='adam', loss='categorical_crossentropy', metrics=['accuracy'])
        return model

    def train(self, x_train, y_train, epochs, batch_size, validation_split, model_save_path):
        checkpoint_callback = ModelCheckpoint(model_save_path, monitor='val_accuracy', save_best_only=True, verbose=1)
        history = self.model.fit(x_train, y_train, epochs=epochs, batch_size=batch_size, validation_split=validation_split, callbacks=[checkpoint_callback])
        return history

class CNNModelPositive:
    def __init__(self, num_words, max_reviews_len, nb_classes):
        self.num_words = num_words
        self.max_reviews_len = max_reviews_len
        self.nb_classes = nb_classes
        self.model = self.build_model()

    def build_model(self):
        model = Sequential()
        model.add(Embedding(self.num_words, 256, input_length=self.max_reviews_len))
        model.add(Conv1D(256, 5, activation='relu'))
        model.add(GlobalMaxPooling1D())
        model.add(Dense(128, activation='relu'))
        model.add(Dense(self.nb_classes, activation='softmax'))
        model.compile(optimizer='adam',
                      loss='categorical_crossentropy',
                      metrics=['accuracy'])
        return model

    def train(self, x_train, y_train, epochs, batch_size, validation_split, model_save_path):
        checkpoint_callback = ModelCheckpoint(model_save_path,
                                              monitor='val_accuracy',
                                              save_best_only=True,
                                              verbose=1)
        history = self.model.fit(x_train,
                                 y_train,
                                 epochs=epochs,
                                 batch_size=batch_size,
                                 validation_split=validation_split,
                                 callbacks=[checkpoint_callback])
        return history

class Plotter:
    @staticmethod
    def plot_history(history, save_path):
        plt.plot(history.history['accuracy'],
                 label='Доля верных ответов на обучающем наборе')
        plt.plot(history.history['val_accuracy'],
                 label='Доля верных ответов на проверочном наборе')
        plt.xlabel('Эпоха обучения')
        plt.ylabel('Доля верных ответов')
        plt.legend()
        plt.savefig(save_path)
        plt.show()

def main():

    tokenizer_path_positive = '../tokenization/tokenizer_positive.pickle'
    class_names_path_positive = '../classification/class_names_positive.txt'
    dataset_path_positive = '../dataset/filtered/filtered_dataset_positive.csv'

    model_save_path_lstm_positive = './model/best_model_lstm_positive.keras'
    plot_save_path_lstm_positive = './graphics/history_lstm_positive.png'

    model_save_path_cnn_positive = './model/best_model_cnn_positive.keras'
    plot_save_path_cnn_positive = './graphics/history_cnn_positive.png'

    model_save_path_gru_positive = './model/best_model_gru_positive.keras'
    plot_save_path_gru_positive = './graphics/history_gru_positive.png'

    tokenizer_path_negative = '../tokenization/tokenizer_negative.pickle'
    class_names_path_negative = '../classification/class_names_negative.txt'
    dataset_path_negative = '../dataset/filtered/filtered_dataset_negative.csv'

    model_save_path_lstm_negative = './model/best_model_lstm_negative.keras'
    plot_save_path_lstm_negative = './graphics/history_lstm_negative.png'

    model_save_path_cnn_negative = './model/best_model_cnn_negative.keras'
    plot_save_path_cnn_negative = './graphics/history_cnn_negative.png'

    model_save_path_gru_negative = './model/best_model_gru_negative.keras'
    plot_save_path_gru_negative = './graphics/history_gru_negative.png'

    data_processor_negative = DataProcessor(tokenizer_path_negative, class_names_path_negative, dataset_path_negative)
    x_train, y_train = data_processor_negative.preprocess_data()

    cnn_model_negative = CNNModelNegative(NUM_WORDS, MAX_REVIEWS_LEN, NB_CLASSES)
    history = cnn_model_negative.train(x_train, y_train, epochs=10, batch_size=128, validation_split=0.1, model_save_path=model_save_path_cnn_negative)
    Plotter.plot_history(history, plot_save_path_cnn_negative)
    print("Training and plotting completed successfully: CNNModelNegative.")

    lstm_model_negative = LSTMModelNegative(NUM_WORDS, MAX_REVIEWS_LEN, NB_CLASSES)
    history = lstm_model_negative.train(x_train, y_train, epochs=10, batch_size=128, validation_split=0.1, model_save_path=model_save_path_lstm_negative)
    Plotter.plot_history(history, plot_save_path_lstm_negative)
    print("Training and plotting completed successfully: LSTMModelNegative.")

    gru_model_negative = GRUModelNegative(NUM_WORDS, MAX_REVIEWS_LEN, NB_CLASSES)
    history = gru_model_negative.train(x_train, y_train, epochs=10, batch_size=128, validation_split=0.1, model_save_path=model_save_path_gru_negative)
    Plotter.plot_history(history, plot_save_path_gru_negative)
    print("Training and plotting completed successfully: GRUModelNegative.")

    data_processor_positive = DataProcessor(tokenizer_path_positive, class_names_path_positive, dataset_path_positive)
    x_train, y_train = data_processor_positive.preprocess_data()

    cnn_model_positive = CNNModelPositive(NUM_WORDS, MAX_REVIEWS_LEN, NB_CLASSES)
    history = cnn_model_positive.train(x_train, y_train, epochs=5, batch_size=128, validation_split=0.1, model_save_path=model_save_path_cnn_positive)
    Plotter.plot_history(history, plot_save_path_cnn_positive)
    print("Training and plotting completed successfully: CNNModelPositive.")

    lstm_model_positive = LSTMModelPositive(NUM_WORDS, MAX_REVIEWS_LEN, NB_CLASSES)
    history = lstm_model_positive.train(x_train, y_train, epochs=5, batch_size=128, validation_split=0.1, model_save_path=model_save_path_lstm_positive)
    Plotter.plot_history(history, plot_save_path_lstm_positive)
    print("Training and plotting completed successfully: LSTMModelPositive.")

    gru_model_positive = GRUModelPositive(NUM_WORDS, MAX_REVIEWS_LEN, NB_CLASSES)
    history = gru_model_positive.train(x_train, y_train, epochs=5, batch_size=128, validation_split=0.1, model_save_path=model_save_path_gru_positive)
    Plotter.plot_history(history, plot_save_path_gru_positive)
    print("Training and plotting completed successfully: GRUModelPositive.")

if __name__ == "__main__":
    main()
