import pandas as pd
from keras.src.legacy.preprocessing.text import Tokenizer
from nltk import word_tokenize, SnowballStemmer
from nltk.corpus import stopwords
import string
import pickle
from pymystem3 import Mystem

# Константы
NUM_WORDS = 10000

class DataLoader:
    def __init__(self, dataset_path):
        self.dataset_path = dataset_path

    def load_data(self):
        return pd.read_csv(self.dataset_path)

class TextProcessor:
    def __init__(self):
        self.stop_words = set(stopwords.words('russian'))
        self.punctuation = set(string.punctuation)
        self.mystem = Mystem()
        self.stemmer = SnowballStemmer("russian")

    def remove_stopwords_and_punctuation(self, reviews):
        filtered_tokens = []
        for review in reviews:
            words = word_tokenize(review)
            filtered_words = [word for word in words if
                              word.lower() not in self.stop_words and '\\n1' and word != "''" and word != '«' and word != '»' and word not in self.punctuation]
            filtered_tokens.extend(filtered_words)
        return filtered_tokens

    def lemmatize_tokens(self, tokens):
        lemmatized_tokens = [self.mystem.lemmatize(token)[0] for token in tokens]
        return lemmatized_tokens

    def stem_tokens(self, tokens):
        stemmed_tokens = [self.stemmer.stem(word) for word in tokens]
        return stemmed_tokens

class TokenizerSaver:
    def __init__(self, num_words):
        self.num_words = num_words
        self.tokenizer = Tokenizer(num_words=self.num_words)

    def fit_on_texts(self, texts):
        self.tokenizer.fit_on_texts(texts)

    def save_tokenizer(self, save_path):
        with open(save_path, 'wb') as handle:
            pickle.dump(self.tokenizer, handle, protocol=pickle.HIGHEST_PROTOCOL)

def process_and_save_tokenizer(dataset_path, save_path):
    # Load data
    data_loader = DataLoader(dataset_path)
    dataset = data_loader.load_data()

    # Process text
    text_processor = TextProcessor()
    reviews = dataset['text']
    filtered_tokens = text_processor.remove_stopwords_and_punctuation(reviews)

    # Tokenize and save
    tokenizer_saver = TokenizerSaver(NUM_WORDS)
    tokenizer_saver.fit_on_texts(filtered_tokens)
    tokenizer_saver.save_tokenizer(save_path)

def main():
    positive_dataset_path = '../dataset/filtered/filtered_dataset_positive.csv'
    negative_dataset_path = '../dataset/filtered/filtered_dataset_negative.csv'
    positive_tokenizer_path = './tokenizer_positive.pickle'
    negative_tokenizer_path = './tokenizer_negative.pickle'

    # Process and save tokenizers
    process_and_save_tokenizer(positive_dataset_path, positive_tokenizer_path)
    process_and_save_tokenizer(negative_dataset_path, negative_tokenizer_path)

if __name__ == "__main__":
    main()
